
#include <stdio.h>
#include <stdlib.h>

extern void	RangeList();

typedef	unsigned char	uChar;
typedef	unsigned short	uShort;

void	Error(msg)
char	*msg;
{
	(void)fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

static FILE	*GIF;

static int	Get()
{
	int	byte=fgetc(GIF);

	if (byte==EOF) Error("byte read error.");

	return byte;
}

static int	GetWord()
{
	int	byte=Get();

	return byte|(Get()<<8);
}

static void	Skip(count)
int		count;
{
/*
	if (fseek(GIF,(long)count,SEEK_CUR)) Error("unexpected end of file.");
*/
	while (--count>=0) (void)Get();
}

static void	GetLUT(rgbSize,BPS,R,G,B,LUT)
int		rgbSize,BPS;
char		*R,*G,*B,*LUT;
{
	int	z,r,g,b,size=1<<BPS;
	double	ratio=255.0/(double)((1<<rgbSize)-1);

	for (z=0; z<size; z++) {
	    r=(int)((double)Get()*ratio+0.5);
	    g=(int)((double)Get()*ratio+0.5);
	    b=(int)((double)Get()*ratio+0.5);
#ifdef	NBG
	    LUT[z]=(!R[r] || !G[g] || !B[b]);
#else
	    LUT[z]=R[r] && G[g] && B[b];
#endif
	}
	for (; z<256; z++) LUT[z]=0;
}

static int	GetBlock(byte)
uChar		*byte;
{
	size_t	size=(size_t)Get();

	if (size!=0 && fread(byte,sizeof(uChar),size,GIF)!=size)
	    Error("data block read error.");

	return (int)size;
}

static void	SkipBlock()
{
	int	count;

	while ((count=Get())!=0) Skip(count);
}

static	void	DecodeImage(x0,y0,Nx,Ny,interlace,codeSize,pv)
int		x0,y0,Nx,Ny,interlace,codeSize;
uChar		**pv;
{
	int	y3,y2,y1,y,x,new,base,last,size,
		Bit=codeSize+1,Clr=1<<(Bit-1),End=Clr+1,Top=End+1,Wid=4097-Top,
		rem=0,num=0,len=Bit,cnt=0,old=Clr,code=Top;
	uChar	*px,*ptr,byte[255],ring[4097-6],Tail[4096-6],
		**py=pv+y0,*tail=Tail-6;
	long	buf=0,mask=(1<<Bit)-1;
	uShort	Head[4096-6],*head=Head-6;

	if (interlace) y3=(y2=(y1=(Ny-1)/8+1)+(Ny-5)/8+1)+(Ny-3)/4+1;

	for (y=0; y<Ny; y++) {
	    px=py[(interlace)?(y<y1)?   y    *8:
			      (y<y2)?4+(y-y1)*8:
			      (y<y3)?2+(y-y2)*4:
				     1+(y-y3)*2:y]+x0;
	    for (x=0; x<Nx; x++) {
		while (rem==0) {
		    while (num<len) {
			if (cnt==0 && (cnt=GetBlock(ptr=byte))==0)
			    Error("missing data block.");

			--cnt; buf|=((long)*(ptr++)<<num); num+=8;
		    }
		    num-=len; new=(int)(buf&mask); buf>>=len; 

		    if (new==End) Error("unexpected end of code.");

		    if (new==Clr) {
			len=Bit; mask=(1<<Bit)-1;
		    }
		    else if (old==Clr) {
			if (new>Clr) Error("bad code after clear code.");

			ring[base=0]=(uChar)(last=new); size=rem=1; code=Top;
		    }
		    else {
			if (code==4096) Error("missing clear code.");

			if (new>code) Error("bad code.");

			if (new==code) {
			    ring[base=(base+Wid-1)%Wid]=(uChar)last; ++size;
			}
			else {
			    base=size=0;
			    for (last=new; last>=Top; last=head[last])
				ring[size++]=tail[last];

			    ring[size++]=(uChar)last;
			}
			rem=size;

			head[code]=(uShort)old; tail[code]=(uChar)last;
			if (code==mask && len<12) {
			    ++len; mask=(mask<<1)|1;
			}
			++code;
		    }
		    old=new;
		}
		*(px++)=ring[(base+(--rem))%Wid];
	    }
	}
	if (rem>0)
	    (void)fputs("data remains (warning).\n",stderr);
	else {
	    while (num<len) {
		if (cnt==0 && (cnt=GetBlock(ptr=byte))==0) {
		    cnt=255; break;
		}
		--cnt; buf|=((long)*(ptr++)<<num); num+=8;
	    }
	    if (cnt!=0 && cnt!=255)
		(void)fputs("too many codes (warning).\n",stderr);
	    else
		if (!(num>=len && (buf&mask)==End) &&
		    !(num>=len-1 && (buf&(mask>>1))==End))
		    (void)fputs("missing end of code (warning).\n",stderr);
	}
	if (cnt!=255) SkipBlock();
}

#define ALLOC(type,size)	(type *)malloc(sizeof(type)*(size_t)(size))

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	R[256],G[256],B[256],gLUT[256],lLUT[256],*LUT,
		*mae="memory allocation error.";
	int	R1,R2,G1,G2,B1,B2,byte,Fx,x1,x2,Fy,y1,y2,
		y,rgbSize,z,x0,y0,x3,Nx,y3,Ny,codeSize,x;
	uChar	**pv;

	if (argc!=5 && argc!=8)
	    Error("usage : gif_area GIF rangeListR rangeListG rangeListB\n"
		  "        gif_area GIF R1 G1 B1 R2 G2 B2");

	if (argc==5) {
	    RangeList(argv[2],(size_t)255,R);
	    RangeList(argv[3],(size_t)255,G);
	    RangeList(argv[4],(size_t)255,B);
	}
	else {
	    if ((R1=atoi(argv[2]))<0 || (R2=atoi(argv[5]))>255 || R1>R2 ||
		(G1=atoi(argv[3]))<0 || (G2=atoi(argv[6]))>255 || G1>G2 ||
		(B1=atoi(argv[4]))<0 || (B2=atoi(argv[7]))>255 || B1>B2)
		Error("bad RGB range.");

	    for (byte=0; byte<=255; byte++) {
		R[byte]=byte>=R1 && byte<=R2;
		G[byte]=byte>=G1 && byte<=G2;
		B[byte]=byte>=B1 && byte<=B2;
	    }
	}
	if (argv[1][0]=='-')
	    GIF=stdin;
	else
	    if ((GIF=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if (Get()!='G' || Get()!='I' || Get()!='F' ||
	    Get()!='8' || ((byte=Get())!='7' && byte!='9') || Get()!='a')
	    Error("bad magic number.");

	Fx=x1=GetWord(); x2=(-1);
	Fy=y1=GetWord(); y2=(-1);

	if ((pv=ALLOC(uChar *,Fy))==NULL) Error(mae);

	for (y=0; y<Fy; y++) if ((pv[y]=ALLOC(uChar,Fx))==NULL) Error(mae);

	byte=Get(); rgbSize=((byte>>4)&7)+1;

	Skip(2);	/* background color index & pixel aspect ratio */

	if (byte&128)
	    GetLUT(rgbSize,(byte&7)+1,R,G,B,gLUT);
	else
	    for (z=0; z<256; z++) gLUT[z]=0;

	while ((byte=Get())!=';')
	    if (byte==',') {
		x0=GetWord();
		y0=GetWord();
		if ((x3=x0+(Nx=GetWord())-1)>=Fx ||
		    (y3=y0+(Ny=GetWord())-1)>=Fy) Error("bad image size.");

		if ((byte=Get())&128)
		    GetLUT(rgbSize,(byte&7)+1,R,G,B,LUT=lLUT);
		else
		    LUT=gLUT;

		if ((codeSize=Get())<2 || codeSize>8) Error("bad code size.");

		if (x1<=x0 && x2>=x3 && y1<=y0 && y2>=y3)
		    SkipBlock();
		else {
		    DecodeImage(x0,y0,Nx,Ny,byte&64,codeSize,pv);

		    for (y=y0; y<=y3; y++)
		    for (x=x0; x<=x3; x++)
			if (LUT[pv[y][x]]) {
			    if (x<x1) x1=x; if (x>x2) x2=x;
			    if (y<y1) y1=y; if (y>y2) y2=y;
			}
		}
	    }
	    else {
		if (byte=='!') Skip(1);	/* control label */

		SkipBlock();
	    }

	if (GIF!=stdin) (void)fclose(GIF);

	if (x1>x2 || y1>y2) Error("no area.");

	(void)printf("%d\t%d\t%d\t%d\n",x1,y1,x2,y2); return 0;
}
