
#include <stdio.h>
#include <stdlib.h>
#include "sgif.h"

typedef	unsigned char	uChar;
typedef	unsigned short	uShort;

static void	Error(msg)
char		*msg;
{
	(void)fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

static FILE	*GIF;

static int	Get()
{
	int	byte=fgetc(GIF);

	if (byte==EOF) Error("byte read error.");

	return byte;
}

static int	GetWord()
{
	int	byte=Get();

	return byte|(Get()<<8);
}

static void	Skip(count)
int		count;
{
/*
	if (fseek(GIF,(long)count,SEEK_CUR)) Error("unexpected end of file.");
*/
	while (--count>=0) (void)Get();
}

static int	GetCM(rgbSize,BPS,r,g,b,R,G,B)
int		rgbSize,BPS,R,G,B;
uChar		*r,*g,*b;
{
	int	z,size=1<<BPS;
	double	ratio=255.0/(double)((1<<rgbSize)-1);

	for (z=0; z<size; z++) {
	    r[z]=(uChar)((double)Get()*ratio+0.5);
	    g[z]=(uChar)((double)Get()*ratio+0.5);
	    b[z]=(uChar)((double)Get()*ratio+0.5);
	}
	for (; z<256; z++) r[z]=g[z]=b[z]=0;

	while (--z>=0) if (r[z]==R && g[z]==G && b[z]==B) break;

	return z;
}

static int	GetBlock(byte)
uChar		*byte;
{
	size_t	size=(size_t)Get();

	if (size!=0 && fread(byte,sizeof(uChar),size,GIF)!=size)
	    Error("data block read error.");

	return (int)size;
}

static void	SkipBlock()
{
	int	count;

	while ((count=Get())!=0) Skip(count);
}

static	void	DecodeImage(x0,y0,Nx,Ny,interlace,codeSize,pv)
int		x0,y0,Nx,Ny,interlace,codeSize;
uChar		**pv;
{
	int	y3,y2,y1,y,x,new,base,last,size,
		Bit=codeSize+1,Clr=1<<(Bit-1),End=Clr+1,Top=End+1,Wid=4097-Top,
		rem=0,num=0,len=Bit,cnt=0,old=Clr,code=Top;
	uChar	*px,*ptr,byte[255],ring[4097-6],Tail[4096-6],
		**py=pv+y0,*tail=Tail-6;
	long	buf=0,mask=(1<<Bit)-1;
	uShort	Head[4096-6],*head=Head-6;

	if (interlace) y3=(y2=(y1=(Ny-1)/8+1)+(Ny-5)/8+1)+(Ny-3)/4+1;

	for (y=0; y<Ny; y++) {
	    px=py[(interlace)?(y<y1)?   y    *8:
			      (y<y2)?4+(y-y1)*8:
			      (y<y3)?2+(y-y2)*4:
				     1+(y-y3)*2:y]+x0;
	    for (x=0; x<Nx; x++) {
		while (rem==0) {
		    while (num<len) {
			if (cnt==0 && (cnt=GetBlock(ptr=byte))==0)
			    Error("missing data block.");

			--cnt; buf|=((long)*(ptr++)<<num); num+=8;
		    }
		    num-=len; new=(int)(buf&mask); buf>>=len; 

		    if (new==End) Error("unexpected end of code.");

		    if (new==Clr) {
			len=Bit; mask=(1<<Bit)-1;
		    }
		    else if (old==Clr) {
			if (new>Clr) Error("bad code after clear code.");

			ring[base=0]=(uChar)(last=new); size=rem=1; code=Top;
		    }
		    else {
			if (code==4096) Error("missing clear code.");

			if (new>code) Error("bad code.");

			if (new==code) {
			    ring[base=(base+Wid-1)%Wid]=(uChar)last; ++size;
			}
			else {
			    base=size=0;
			    for (last=new; last>=Top; last=head[last])
				ring[size++]=tail[last];

			    ring[size++]=(uChar)last;
			}
			rem=size;

			head[code]=(uShort)old; tail[code]=(uChar)last;
			if (code==mask && len<12) {
			    ++len; mask=(mask<<1)|1;
			}
			++code;
		    }
		    old=new;
		}
		*(px++)=ring[(base+(--rem))%Wid];
	    }
	}
	if (rem>0)
	    (void)fputs("data remains (warning).\n",stderr);
	else {
	    while (num<len) {
		if (cnt==0 && (cnt=GetBlock(ptr=byte))==0) {
		    cnt=255; break;
		}
		--cnt; buf|=((long)*(ptr++)<<num); num+=8;
	    }
	    if (cnt!=0 && cnt!=255)
		(void)fputs("too many codes (warning).\n",stderr);
	    else
		if (!(num>=len && (buf&mask)==End) &&
		    !(num>=len-1 && (buf&(mask>>1))==End))
		    (void)fputs("missing end of code (warning).\n",stderr);
	}
	if (cnt!=255) SkipBlock();
}

#define ALLOC(type,size)	(type *)malloc(sizeof(type)*(size_t)(size))

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	R,G,B,byte,Fx,Fy,x1,y1,x2,Ix,y2,Iy,
		y,rgbSize,Z,z,x0,y0,x3,Nx,y3,Ny,l,codeSize,x;
	uChar	**pvR,**pvG,**pvB,**pv,gR[256],gG[256],gB[256],*r,*g,*b,
		lR[256],lG[256],lB[256],**Ry,**Gy,**By,*Rx,*Gx,*Bx,**Y,*X;
	char	*mae="memory allocation error.";
	FILE	*file;

	if (argc!=10) Error("usage : gif_trim orgGIF x1 y1 x2 y2 R G B newGIF");

	if ((R=atoi(argv[6]))<0 || R>255 ||
	    (G=atoi(argv[7]))<0 || G>255 ||
	    (B=atoi(argv[8]))<0 || B>255) Error("bad RGB value.");

	if (argv[1][0]=='-')
	    GIF=stdin;
	else
	    if ((GIF=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if (Get()!='G' || Get()!='I' || Get()!='F' ||
	    Get()!='8' || ((byte=Get())!='7' && byte!='9') || Get()!='a')
	    Error("bad magic number.");

	Fx=GetWord();
	Fy=GetWord();

	if ((x1=atoi(argv[2]))<0 ||
	    (y1=atoi(argv[3]))<0 ||
	    (x2=atoi(argv[4]))>=Fx || (Ix=x2-x1+1)<1 ||
	    (y2=atoi(argv[5]))>=Fy || (Iy=y2-y1+1)<1) Error("bad image area.");

	if ((pvR=ALLOC(uChar *,Fy))==NULL ||
	    (pvG=ALLOC(uChar *,Fy))==NULL ||
	    (pvB=ALLOC(uChar *,Fy))==NULL) Error(mae);

	for (y=0; y<Fy; y++)
	    if ((pvR[y]=ALLOC(uChar,Fx))==NULL ||
		(pvG[y]=ALLOC(uChar,Fx))==NULL ||
		(pvB[y]=ALLOC(uChar,Fx))==NULL) Error(mae);

	pv=pvB;

	if (argv[9][0]=='-')
	    file=stdout;
	else
	    if ((file=fopen(argv[9],"wb"))==NULL) Error("file not created.");

	byte=Get(); rgbSize=((byte>>4)&7)+1;

	Skip(2);	/* background color index & pixel aspect ratio */

	if (byte&128)
	    Z=GetCM(rgbSize,(byte&7)+1,r=gR,g=gG,b=gB,R,G,B);
	else {
	    Z=(-1); r=NULL; g=NULL; b=NULL;

	    for (z=0; z<256; z++) gR[z]=gG[z]=gB[z]=StoreGIF_Gray[z];
	}
	InitStoreGIF(file,Ix,Iy,r,g,b);

	while ((byte=Get())!=';')
	    if (byte==',') {
		x0=GetWord();
		y0=GetWord();
		if ((x3=x0+(Nx=GetWord())-1)>=Fx ||
		    (y3=y0+(Ny=GetWord())-1)>=Fy) Error("bad image size.");

		if (l=(byte=Get())&128)
		    z=GetCM(rgbSize,(byte&7)+1,r=lR,g=lG,b=lB,R,G,B);
		else {
		    z=Z; r=gR; g=gG; b=gB;
		}
		if ((codeSize=Get())<2 || codeSize>8) Error("bad code size.");

		DecodeImage(x0,y0,Nx,Ny,byte&64,codeSize,pv);

		if (z==(-1) && (x1<x0 || x2>x3 || y1<y0 || y2>y3)) {
		    for (Ry=pvR,
			 Gy=pvG,
			 By=pvB, y=y1; y<=y2; y++, Ry++, Gy++, By++)
		    for (Rx=(*Ry),
			 Gx=(*Gy),
			 Bx=(*By), x=x1; x<=x2; x++, Rx++, Gx++, Bx++)
			if (x>=x0 && x<=x3 && y>=y0 && y<=y3) {
			    z=pv[y][x]; *Rx=r[z]; *Gx=g[z]; *Bx=b[z];
			}
			else {
			    *Rx=(uChar)R; *Gx=(uChar)G; *Bx=(uChar)B;
			}

		    StoreGIF_RGB(file,0,0,Ix,Iy,pvR,pvG,pvB);
		}
		else {
		    for (Y=pv, y=y1; y<=y2; y++, Y++)
		    for (X=(*Y), x=x1; x<=x2; x++, X++)
			*X=(x>=x0 && x<=x3 && y>=y0 && y<=y3)?pv[y][x]:(uChar)z;

		    if (!l) r=g=b=NULL;

		    StoreGIF(file,0,0,Ix,Iy,r,g,b,pv);
		}
	    }
	    else {
		if (byte=='!') Skip(1);	/* control label */

		SkipBlock();
	    }

	(void)TermStoreGIF(file); if (file!=stdout) (void)fclose(file);

	if (GIF!=stdin) (void)fclose(GIF);

	return 0;
}
