
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "csi.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

static int	X,Y,Z,facets=0;

#ifdef	STL_BINARY
static size_t	fWriteReverse(ptr,size,items,file)
char		*ptr;
size_t		size,items;
FILE		*file;
{
	char	*p,*h,*t,c;
	size_t	i,s=size-1;

	for (p=ptr, i=0; i<items; i++, p+=size)
	    for (h=p, t=p+s; h<t; h++, t--) {
		c=(*h); *h=(*t); *t=c;
	    }

	return fwrite(ptr,size,items,file);
}

static FILE	*file;
static size_t	(*fWrite)();

static unsigned short	flag=0;

static void	PrintFacet(dx,dy,dz,x0,y0,z0,x1,y1,z1,x2,y2,z2)
int		dx,dy,dz,x0,y0,z0,x1,y1,z1,x2,y2,z2;
{
	float	xyz[12];

	xyz[0]=(float)   dx ; xyz[ 1]=(float)   dy ; xyz[ 2]=(float)   dz ;
	xyz[3]=(float)(X+x0); xyz[ 4]=(float)(Y+y0); xyz[ 5]=(float)(Z+z0);
	xyz[6]=(float)(X+x1); xyz[ 7]=(float)(Y+y1); xyz[ 8]=(float)(Z+z1);
	xyz[9]=(float)(X+x2); xyz[10]=(float)(Y+y2); xyz[11]=(float)(Z+z2);

	if ((*fWrite)((char *)xyz,sizeof(*xyz),12,file)!=12 ||
	    (*fWrite)((char *)&flag,sizeof(flag),1,file)!=1)
	    Error("STL data write error.");

	++facets;
}
#else
static void	PrintFacet(dx,dy,dz,x0,y0,z0,x1,y1,z1,x2,y2,z2)
int		dx,dy,dz,x0,y0,z0,x1,y1,z1,x2,y2,z2;
{
	printf("facet normal %d %d %d\n",dx,dy,dz);
	puts("outer loop");
	printf("vertex %d %d %d\n",X+x0,Y+y0,Z+z0);
	printf("vertex %d %d %d\n",X+x1,Y+y1,Z+z1);
	printf("vertex %d %d %d\n",X+x2,Y+y2,Z+z2);
	puts("endloop\nendfacet");

	++facets;
}
#endif

static void	PrintMinusZ(x0,y0,x1,y1,x2,y2)
int		x0,y0,x1,y1,x2,y2;
{
	PrintFacet(0,0,-1,x2,y2,0,x1,y1,0,x0,y0,0);
}

static void	PrintPlusZ(x0,y0,x1,y1,x2,y2)
int		x0,y0,x1,y1,x2,y2;
{
	PrintFacet(0,0, 1,x0,y0,1,x1,y1,1,x2,y2,1);
}

static void	PrintMinusX(y0,z0,y1,z1,y2,z2)
int		y0,z0,y1,z1,y2,z2;
{
	PrintFacet(-1,0,0,0,y2,z2,0,y1,z1,0,y0,z0);
}

static void	PrintPlusX(y0,z0,y1,z1,y2,z2)
int		y0,z0,y1,z1,y2,z2;
{
	PrintFacet( 1,0,0,1,y0,z0,1,y1,z1,1,y2,z2);
}

static void	PrintMinusY(z0,x0,z1,x1,z2,x2)
int		z0,x0,z1,x1,z2,x2;
{
	PrintFacet(0,-1,0,x2,0,z2,x1,0,z1,x0,0,z0);
}

static void	PrintPlusY(z0,x0,z1,x1,z2,x2)
int		z0,x0,z1,x1,z2,x2;
{
	PrintFacet(0, 1,0,x0,1,z0,x1,1,z1,x2,1,z2);
}

typedef struct List {
		int		x,y;
		struct List	*next;
	} List;

static List	*Alloc(x,y,next)
int		x,y;
List		*next;
{
	List	*list=(List *)malloc(sizeof(*list));

	if (list==NULL) Error("list memory allocation error.");

	list->x=x; list->y=y; list->next=next; return list;
}

#define ALLOC(type,size)	(type *)malloc(sizeof(type)*(size_t)(size))

static void	(*Print)();

#define PRINT(a,b,c)	(*Print)(a->x,a->y,b->x,b->y,c->x,c->y)

static void	Flush(list)
List		*list;
{
	int	idx,oSide,nSide,x0,y0,x1,y1,x2,y2,cnt=0;
	List	**array,**head,**tail,*oList,*nList,*stack=list;

	while ((list=list->next)!=NULL) ++cnt;

	if ((array=ALLOC(List *,cnt))==NULL)
	    Error("array memory allocation error.");

	list=stack;
	for (idx=0; idx<cnt; idx++) {
	    array[idx]=list; list=list->next;
	}
	free(list);

	stack->next=NULL;

	head=array+1; tail=array+cnt-1;
	if ((*head)->y>(*tail)->y) {
	    oSide=0; oList=(*(head++));
	}
	else {
	    oSide=1; oList=(*(tail--));
	}
	while (head!=tail) {
	    if ((*head)->y>(*tail)->y) {
		nSide=0; nList=(*(head++));
	    }
	    else {
		nSide=1; nList=(*(tail--));
	    }
	    if (oSide==nSide) {
		x2=oList->x-(x0=nList->x); y2=oList->y-(y0=nList->y);
		if (oSide==0)
		    do {
			x1=x2; y1=y2;
			if (y1*(x2=stack->x-x0)<=x1*(y2=stack->y-y0)) break;

			PRINT(stack,oList,nList); oList=stack;
		    } while ((stack=stack->next)!=NULL);
		else /* if (oSide==1) */
		    do {
			x1=x2; y1=y2;
			if (y1*(x2=stack->x-x0)>=x1*(y2=stack->y-y0)) break;

			PRINT(stack,nList,oList); oList=stack;
		    } while ((stack=stack->next)!=NULL);
	    }
	    else {
		list=oList;
		if (oSide==0)
		    do {
			PRINT(stack,oList,nList); oList=stack;
		    } while ((stack=stack->next)!=NULL);
		else /* if (oSide==1) */
		    do {
			PRINT(stack,nList,oList); oList=stack;
		    } while ((stack=stack->next)!=NULL);

		oList=list;
	    }
	    oList->next=stack; stack=oList; oList=nList; oSide=nSide;
	}
	nList=(*head);
	if (oSide==0)
	    do {
		PRINT(stack,oList,nList); oList=stack;
	    } while ((stack=stack->next)!=NULL);
	else /* if (oSide==1) */
	    do {
		PRINT(stack,nList,oList); oList=stack;
	    } while ((stack=stack->next)!=NULL);

	for (idx=0; idx<cnt; idx++) free(array[idx]); free(array);
}

typedef unsigned char	Byte;

static List	*hori,**vert,**help,**ptr;

static void	Corner(x,y,a,b,c,d)
int		x,y;
Byte		a,b,c,d;
{
	List	*temp,*next;

#define VERT	vert[x]
#define HELP	help[x]

switch(((a==1)?8:0)|((b==1)?4:0)|((c==1)?2:0)|((d==1)?1:0)) {
    case 0x1:
	VERT=Alloc(x,y+1,Alloc(x,y,hori=Alloc(x+1,y,NULL)));
	ptr=(&HELP);
	break;
    case 0x2:
	VERT->y=y;
	next=HELP->next;
	if (HELP->x==next->x && HELP->y==next->y) {
	    hori=Alloc(x+1,y,Alloc(x,y,next));
	    HELP->next=Alloc(x,y,NULL); Flush(VERT);
	}
	else
	    hori=Alloc(x+1,y,VERT);
	break;
    case 0x3:
	if ((a==3 && b!=3) || (a!=3 && b==3)) {
	    VERT->y=y; VERT=Alloc(x,y+1,VERT);
	}
	ptr=(&HELP);
	break;
    case 0x4:
	hori->x=x; hori->next=VERT=Alloc(x,y+1,NULL);
	*ptr=hori;
	break;
    case 0x5:
	if ((a==3 && c!=3) || (a!=3 && c==3)) {
	    hori->x=x; hori=hori->next=Alloc(x+1,y,NULL);
	}
	break;
    case 0x6:
	VERT->y=y;
	next=HELP->next;
	if (HELP->x==next->x && HELP->y==next->y) {
	    temp=Alloc(x+1,y,Alloc(x,y,next));
	    HELP->next=Alloc(x,y,NULL); Flush(VERT);
	}
	else
	    temp=Alloc(x+1,y,VERT);
	hori->x=x; hori->next=VERT=Alloc(x,y+1,NULL);
	*ptr=hori;
	hori=temp;
	break;
    case 0x7:
	hori->x=x; VERT->y=y;
	next=HELP->next;
	if (HELP->x==next->x && HELP->y==next->y) {
	    hori->next=Alloc(x,y,next);
	    HELP->next=Alloc(x,y,NULL); Flush(VERT);
	}
	else
	    hori->next=VERT;
	*ptr=hori;
	break;
    case 0x8:
	VERT->y=y; hori->x=x; Flush(hori);
	break;
    case 0x9:
	VERT->y=y; hori->x=x; Flush(hori);
	VERT=Alloc(x,y+1,Alloc(x,y,hori=Alloc(x+1,y,NULL)));
	ptr=(&HELP);
	break;
    case 0xa:
	if ((b==3 && d!=3) || (b!=3 && d==3)) {
	    hori->x=x; hori=Alloc(x+1,y,hori);
	}
	break;
    case 0xb:
	hori->x=x; VERT=Alloc(x,y+1,hori);
	ptr=(&HELP); HELP=hori;
	break;
    case 0xc:
	if ((c==3 && d!=3) || (c!=3 && d==3)) {
	    VERT->y=y; VERT=VERT->next=Alloc(x,y+1,NULL);
	}
	break;
    case 0xd:
	VERT->y=y;
	temp=(*ptr); next=temp->next;
	if (temp->x==next->x && temp->y==next->y) {
	    Flush(Alloc(x,y,next));
	    temp->next=Alloc(x,y,hori=Alloc(x+1,y,NULL));
	}
	else
	    VERT->next=hori=Alloc(x+1,y,NULL);
	break;
    case 0xe:
	temp=(*ptr); next=temp->next;
	if (temp->x==next->x && temp->y==next->y)
	    hori=Alloc(x+1,y,Alloc(x,y,next));
	else
	    hori=Alloc(x+1,y,Alloc(x,y,Alloc(temp->x,temp->y,next)));
	*ptr=temp->next=Alloc(x,y,VERT=Alloc(x,y+1,NULL));
	break;
}}

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	lo,up,Nx,Ny,Nz,x1,y1,z1,x2,y2,z2,x,y,z,nx,ny,nz,N,n,x0,y0,z0;
	char	**path,
		*imae="image memory allocation error.",
		*wmae="work memory allocation error.";
	Byte	***image,**slice,bit,**face;
	Cell	**cell;
	long	pels=0L,faces=0L;

#ifdef	STL_BINARY
	if (argc!=6 && argc!=9)
	    Error("usage : si_stl_B directory name lower upper {R G B} STL");

	if (argc==9)
{
	int	R=atoi(argv[5]),
		G=atoi(argv[6]),
		B=atoi(argv[7]);

	if (R<0 || R>255 || G<0 || G>255 || B<0 || B>255) Error("bad color.");

	flag=(unsigned short)(32768|((R&248)<<7)|((G&248)<<2)|((B&248)>>3));
}
#else
	if (argc!=5) Error("usage : si_stl_A directory name lower upper");
#endif
	if ((lo=atoi(argv[3]))<0 ||
	    (up=atoi(argv[4]))<lo) Error("bad pixel value range.");

	path=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);

	fprintf(stderr,"%d\t%d\t%d\n",Nx,Ny,Nz);

	if ((image=ALLOC(Byte **,(Nz+7)/8))==NULL) Error(imae);

	x1=Nx-1;
	y1=Ny-1;
	z1=Nz-1; x2=y2=z2=0;
	for (z=0; z<Nz; z++) {
	    if (z%8==0) {
		if ((slice=ALLOC(Byte *,Ny))==NULL) Error(imae);

		for (y=0; y<Ny; y++) {
		    if ((slice[y]=ALLOC(Byte,Nx))==NULL) Error(imae);

		    for (x=0; x<Nx; x++) slice[y][x]=0;
		}
		image[z/8]=slice; bit=1;
	    }
	    else
		bit<<=1;

	    ReadImageFile(path[z],&nx,&ny,NULL,&cell,NULL);

	    for (y=0; y<ny; y++) {
		for (x=0; x<nx; x++)
		    if ((int)cell[y][x]>=lo && (int)cell[y][x]<=up) {
			slice[y][x]|=bit;
			if (x<x1) x1=x; if (x>x2) x2=x;
			if (y<y1) y1=y; if (y>y2) y2=y;
			if (z<z1) z1=z; if (z>z2) z2=z; ++pels;
		    }

		free(cell[y]);
	    }
	    free(cell);
	}
	fprintf(stderr,"%ld\t%d\t%d\t%d\t%d\t%d\t%d\n",pels,x1,y1,z1,x2,y2,z2);

	if (pels==0L) Error("no object exists.");

	nx=x2-x1+1;
	ny=y2-y1+1;
	nz=z2-z1+1; N=(nx>ny)?(nx>nz)?nx:nz:(ny>nz)?ny:nz;

	if ((face=ALLOC(Byte *,N+1))==NULL ||
	    (vert=ALLOC(List *,N+1))==NULL ||
	    (help=ALLOC(List *,N))==NULL) Error(wmae);

	++face;

	for (n=(-1); n<N; n++) {
	    if ((face[n]=ALLOC(Byte,N+1))==NULL) Error(wmae);

	    ++face[n]; face[-1][n]=face[n][-1]=0;
	}

#ifdef	STL_BINARY
	if ((file=fopen(argv[argc-1],"wb"))==NULL)
	    Error("STL file not stored.");
{
	int	i;
	short	s=0x1234;
	char	*c=(char *)&s;

	for (i=80+sizeof(facets); i>0; i--)
	    if (fputc(0,file)!=0) Error("STL header data write error.");

	fWrite=(*c==(s&0xff))?fwrite:fWriteReverse;
}
#else
	puts("solid");
#endif

X=x1; Y=y1;

    Print=PrintMinusZ;

    for (y=0; y<ny; y++) for (x=0; x<nx; x++) face[y][x]=0;

    for (Z=z1; Z<=z2; Z++) {
	slice=image[Z/8]; bit=(Byte)(1<<(Z%8));
	for (y0=y1, y=0; y<ny; y++, y0++) {
	    for (x0=x1, x=0; x<nx; x++, x0++) {
		if (slice[y0][x0]&bit)
		    if (face[y][x]&1)
			face[y][x]=3;
		    else {
			face[y][x]=1; ++faces;
		    }
		else
		    face[y][x]=(face[y][x]&1)?2:0;

		Corner(x,y,face[y-1][x-1],face[y][x-1],face[y-1][x],face[y][x]);
	    }
	    Corner(x,y,face[y-1][x-1],face[y][x-1],0,0);
	}
	for (x=0; x<nx; x++) Corner(x,y,face[y-1][x-1],0,face[y-1][x],0);

	Corner(x,y,face[y-1][x-1],0,0,0);
    }

    Print=PrintPlusZ;

    for (y=0; y<ny; y++) for (x=0; x<nx; x++) face[y][x]=0;

    for (Z=z2; Z>=z1; Z--) {
	slice=image[Z/8]; bit=(Byte)(1<<(Z%8));
	for (y0=y1, y=0; y<ny; y++, y0++) {
	    for (x0=x1, x=0; x<nx; x++, x0++) {
		if (slice[y0][x0]&bit)
		    if (face[y][x]&1)
			face[y][x]=3;
		    else {
			face[y][x]=1; ++faces;
		    }
		else
		    face[y][x]=(face[y][x]&1)?2:0;

		Corner(x,y,face[y-1][x-1],face[y][x-1],face[y-1][x],face[y][x]);
	    }
	    Corner(x,y,face[y-1][x-1],face[y][x-1],0,0);
	}
	for (x=0; x<nx; x++) Corner(x,y,face[y-1][x-1],0,face[y-1][x],0);

	Corner(x,y,face[y-1][x-1],0,0,0);
    }

Y=y1; Z=z1;

    Print=PrintMinusX;

    for (z=0; z<nz; z++) for (y=0; y<ny; y++) face[z][y]=0;

    for (X=x1; X<=x2; X++) {
	for (z0=z1, z=0; z<nz; z++, z0++) {
	    slice=image[z0/8]; bit=(Byte)(1<<(z0%8));
	    for (y0=y1, y=0; y<ny; y++, y0++) {
		if (slice[y0][X]&bit)
		    if (face[z][y]&1)
			face[z][y]=3;
		    else {
			face[z][y]=1; ++faces;
		    }
		else
		    face[z][y]=(face[z][y]&1)?2:0;

		Corner(y,z,face[z-1][y-1],face[z][y-1],face[z-1][y],face[z][y]);
	    }
	    Corner(y,z,face[z-1][y-1],face[z][y-1],0,0);
	}
	for (y=0; y<ny; y++) Corner(y,z,face[z-1][y-1],0,face[z-1][y],0);

	Corner(y,z,face[z-1][y-1],0,0,0);
    }

    Print=PrintPlusX;

    for (z=0; z<nz; z++) for (y=0; y<ny; y++) face[z][y]=0;

    for (X=x2; X>=x1; X--) {
	for (z0=z1, z=0; z<nz; z++, z0++) {
	    slice=image[z0/8]; bit=(Byte)(1<<(z0%8));
	    for (y0=y1, y=0; y<ny; y++, y0++) {
		if (slice[y0][X]&bit)
		    if (face[z][y]&1)
			face[z][y]=3;
		    else {
			face[z][y]=1; ++faces;
		    }
		else
		    face[z][y]=(face[z][y]&1)?2:0;

		Corner(y,z,face[z-1][y-1],face[z][y-1],face[z-1][y],face[z][y]);
	    }
	    Corner(y,z,face[z-1][y-1],face[z][y-1],0,0);
	}
	for (y=0; y<ny; y++) Corner(y,z,face[z-1][y-1],0,face[z-1][y],0);

	Corner(y,z,face[z-1][y-1],0,0,0);
    }

Z=z1; X=x1;

    Print=PrintMinusY;

    for (x=0; x<nx; x++) for (z=0; z<nz; z++) face[x][z]=0;

    for (Y=y1; Y<=y2; Y++) {
	for (x0=x1, x=0; x<nx; x++, x0++) {
	    for (z0=z1, z=0; z<nz; z++, z0++) {
		if (image[z0/8][Y][x0]&(Byte)(1<<(z0%8)))
		    if (face[x][z]&1)
			face[x][z]=3;
		    else {
			face[x][z]=1; ++faces;
		    }
		else
		    face[x][z]=(face[x][z]&1)?2:0;

		Corner(z,x,face[x-1][z-1],face[x][z-1],face[x-1][z],face[x][z]);
	    }
	    Corner(z,x,face[x-1][z-1],face[x][z-1],0,0);
	}
	for (z=0; z<nz; z++) Corner(z,x,face[x-1][z-1],0,face[x-1][z],0);

	Corner(z,x,face[x-1][z-1],0,0,0);
    }

    Print=PrintPlusY;

    for (x=0; x<nx; x++) for (z=0; z<nz; z++) face[x][z]=0;

    for (Y=y2; Y>=y1; Y--) {
	for (x0=x1, x=0; x<nx; x++, x0++) {
	    for (z0=z1, z=0; z<nz; z++, z0++) {
		if (image[z0/8][Y][x0]&(Byte)(1<<(z0%8)))
		    if (face[x][z]&1)
			face[x][z]=3;
		    else {
			face[x][z]=1; ++faces;
		    }
		else
		    face[x][z]=(face[x][z]&1)?2:0;

		Corner(z,x,face[x-1][z-1],face[x][z-1],face[x-1][z],face[x][z]);
	    }
	    Corner(z,x,face[x-1][z-1],face[x][z-1],0,0);
	}
	for (z=0; z<nz; z++) Corner(z,x,face[x-1][z-1],0,face[x-1][z],0);

	Corner(z,x,face[x-1][z-1],0,0,0);
    }

	fprintf(stderr,"%ld\t%d\n",faces,facets);

#ifdef	STL_BINARY
	if ((file=freopen(argv[argc-1],"rb+",file))==NULL ||
	    fseek(file,80L,SEEK_SET) ||
	    (*fWrite)(&facets,sizeof(facets),1,file)!=1)
	    Error("STL trailer data write error.");

	fclose(file);
#else
	puts("endsolid");
#endif
	return 0;
}
