
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "stl.h"
#include "xyz.h"
#include "sgif.h"

#define gamma	Gamma	/* to avoid duplication error */

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

typedef unsigned char	byte;

static byte	**cell;
static int	Ny1;

#define GRAY	(byte)(bias+(255.0-bias)*pow(shade,1.0/gamma)+0.5)

#ifdef	SS
static double	gamma,bias;

static void	Shade(x,y,shade)
int		x,y;
double		shade;
{
	cell[Ny1-y][x]=GRAY;
}
#endif

static byte	pv;

static void	Plot(x,y)
int		x,y;
{
	cell[Ny1-y][x]=pv;
}

static double	Distance(x,y,z,D)
double		x,y,z,D;
{
	double	d=sqrt(x*x+y*y+z*z);

	return (d>D)?d:D;
}

static void	Rotate(C,x,y,z,X,Y,Z)
double		C[3][3],x,y,z,*X,*Y,*Z;
{
	*X=C[0][0]*x+C[0][1]*y+C[0][2]*z;
	*Y=C[1][0]*x+C[1][1]*y+C[1][2]*z;
	*Z=C[2][0]*x+C[2][1]*y+C[2][2]*z;
}

#define STR_LEN		256
#define RADIAN		(M_PI/180.0)
#define MA(num,var)	malloc((size_t)(num)*sizeof(*(var)))

int	main(argc,argv)
int	argc;
char	**argv;
{
	double	xO,yO,zO,lon,lat,scale,
		D,H,V,t0,dt,l,p,cl,sl,cp,sp,A[3][3],X,Y,Z,R,r,d,
		t,ct,st,B[2][3],C[3][3],X0,Y0,Z0,X1,Y1,Z1,X2,Y2,Z2;
	int	bg,fg,idx,Nx,Ny,y,cnt,x;
	FILE	*file;
	Facet	*facet,*top=NULL,**ptr=(&top);
	char	line[STR_LEN],str[5][STR_LEN],
		*cmae="cell memory allocation error.";
#ifndef	SS
	double	gamma,bias,shade;
#endif

	if (argc<13)
	Error(
"usage : stl_ar_GIF STL_files xO yO zO lon lat scale gamma bias bg fg GIF"
	);

	xO=atof(argv[argc-11]);
	yO=atof(argv[argc-10]);
	zO=atof(argv[argc- 9]);
	lon=atof(argv[argc-8])*RADIAN;
	lat=atof(argv[argc-7])*RADIAN;

	if ((scale=atof(argv[argc-6]))<=0.0) Error("bad scale.");
	if ((gamma=atof(argv[argc-5]))<=0.0) Error("bad gamma.");
	if ((bias=atof(argv[argc-4]))<0.0 || bias>255.0) Error("bad bias.");
	if ((bg=atoi(argv[argc-3]))<0 || bg>255) Error("bad background.");
	if ((fg=atoi(argv[argc-2]))>255) Error("bad foreground.");

	for (idx=1; idx<argc-11; idx++) {
	    if ((file=fopen(argv[idx],"rb"))==NULL) {
		fprintf(stderr,"%s : file not found.\n",argv[idx]); exit(1);
	    }
	    if ((*ptr=ReadFacet(file))==NULL) {
		fprintf(stderr,"%s : file read error.\n",argv[idx]); exit(1);
	    }
	    fclose(file);

	    for (facet=(*ptr); facet->next!=NULL; facet=facet->next) ;

	    ptr=(&(facet->next));
	}
	D=0.0;
	for (facet=top; facet!=NULL; facet=facet->next) {
	    D=Distance(facet->x0-xO,facet->y0-yO,facet->z0-zO,D);
	    D=Distance(facet->x1-xO,facet->y1-yO,facet->z1-zO,D);
	    D=Distance(facet->x2-xO,facet->y2-yO,facet->z2-zO,D);
	}

	XYZWidth(lon,lat,-D,-D,-D,D,D,D,&H,&V);
	Nx=(int)ceil(scale*H);
	Ny=(int)ceil(scale*V); Ny1=Ny-1;

	if ((cell=(byte **)MA(Ny,cell))==NULL) Error(cmae);

	for (y=0; y<Ny; y++)
	    if ((cell[y]=(byte *)MA(Nx,*cell))==NULL) Error(cmae);

	if (argv[argc-1][0]=='-')
	    file=stdout;
	else
	    if ((file=fopen(argv[argc-1],"wb"))==NULL)
		Error("GIF file not created.");

	InitStoreGIF(file,Nx,Ny,StoreGIF_Gray,StoreGIF_Gray,StoreGIF_Gray);

	while (gets(line)!=NULL)
{
	switch(sscanf(line,"%s %s %s %s %s",
			   str[0],str[1],str[2],str[3],str[4])) {
	    case 3:
		t0=atof(str[2]); dt=0.0; cnt=1; break;
	    case 5:
		t0=atof(str[2]); dt=atof(str[3]); cnt=atoi(str[4]); break;
	    default:
		continue;
	}
	l=atof(str[0])*RADIAN; cl=cos(l); sl=sin(l);
	p=atof(str[1])*RADIAN; cp=cos(p); sp=sin(p);
	A[0][0]=(-sl); A[0][1]=(-cl*sp); A[0][2]=cl*cp;
	A[1][0]=  cl ; A[1][1]=(-sl*sp); A[1][2]=sl*cp;
	A[2][0]= 0.0 ; A[2][1]=     cp ; A[2][2]=   sp;

	R=fabs(X=A[0][2]);
	if ((r=fabs(Y=A[1][2]))>R) R=r;
	if ((r=fabs(Z=A[2][2]))>R) R=r; d=D/R; X*=d; Y*=d; Z*=d;

	for (idx=0; idx<cnt; idx++) {
	    t=(t0+dt*(double)idx)*RADIAN; ct=cos(t); st=sin(t);
	    B[0][0]=ct*A[0][0]-st*A[0][1];
	    B[0][1]=ct*A[1][0]-st*A[1][1];
	    B[0][2]=ct*A[2][0]-st*A[2][1];
	    B[1][0]=st*A[0][0]+ct*A[0][1];
	    B[1][1]=st*A[1][0]+ct*A[1][1];
	    B[1][2]=st*A[2][0]+ct*A[2][1];
	    C[0][0]=A[0][0]*B[0][0]+A[0][1]*B[1][0]+A[0][2]*A[0][2];
	    C[0][1]=A[0][0]*B[0][1]+A[0][1]*B[1][1]+A[0][2]*A[1][2];
	    C[0][2]=A[0][0]*B[0][2]+A[0][1]*B[1][2]+A[0][2]*A[2][2];
	    C[1][0]=A[1][0]*B[0][0]+A[1][1]*B[1][0]+A[1][2]*A[0][2];
	    C[1][1]=A[1][0]*B[0][1]+A[1][1]*B[1][1]+A[1][2]*A[1][2];
	    C[1][2]=A[1][0]*B[0][2]+A[1][1]*B[1][2]+A[1][2]*A[2][2];
	    C[2][0]=A[2][0]*B[0][0]+A[2][1]*B[1][0]+A[2][2]*A[0][2];
	    C[2][1]=A[2][0]*B[0][1]+A[2][1]*B[1][1]+A[2][2]*A[1][2];
	    C[2][2]=A[2][0]*B[0][2]+A[2][1]*B[1][2]+A[2][2]*A[2][2];

	    for (y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++) cell[y][x]=(byte)bg;

	    if (!XYZInit(Nx,Ny,lon,lat,-D,-D,-D,D,D,D))
		Error("rendering memory allocation error.");

	    for (facet=top; facet!=NULL; facet=facet->next) {
		Rotate(C,facet->x0-xO,facet->y0-yO,facet->z0-zO,&X0,&Y0,&Z0);
		Rotate(C,facet->x1-xO,facet->y1-yO,facet->z1-zO,&X1,&Y1,&Z1);
		Rotate(C,facet->x2-xO,facet->y2-yO,facet->z2-zO,&X2,&Y2,&Z2);
#ifndef	SS
		if ((shade=XYZSide(lon,lat,X0,Y0,Z0,X1,Y1,Z1,X2,Y2,Z2))>=0.0) {
		    pv=GRAY; XYZFill(X0,Y0,Z0,X1,Y1,Z1,X2,Y2,Z2,Plot);
		}
	    }
#else
		if (XYZSide(lon,lat,X0,Y0,Z0,X1,Y1,Z1,X2,Y2,Z2)>=0.0)
		    XYZFill(X0,Y0,Z0,X1,Y1,Z1,X2,Y2,Z2,NULL);
	    }
	    XYZShade(lon,lat,Shade);
#endif
	    if (fg>=0) {
		pv=(byte)fg;
		XYZDraw(-D,-D,-D, D,-D,-D,Plot);
		XYZDraw( D,-D,-D, D, D,-D,Plot);
		XYZDraw( D, D,-D,-D, D,-D,Plot);
		XYZDraw(-D, D,-D,-D,-D,-D,Plot);
		XYZDraw(-D,-D, D, D,-D, D,Plot);
		XYZDraw( D,-D, D, D, D, D,Plot);
		XYZDraw( D, D, D,-D, D, D,Plot);
		XYZDraw(-D, D, D,-D,-D, D,Plot);
		XYZDraw(-D,-D,-D,-D,-D, D,Plot);
		XYZDraw( D,-D,-D, D,-D, D,Plot);
		XYZDraw( D, D,-D, D, D, D,Plot);
		XYZDraw(-D, D,-D,-D, D, D,Plot); XYZDraw(-X,-Y,-Z,X,Y,Z,Plot);
	    }
	    XYZTerm();

	    StoreGIF(file,0,0,Nx,Ny,NULL,NULL,NULL,cell);
	}
}
	TermStoreGIF(file);

	if (file!=stdout) fclose(file); return 0;
}
