
#ifdef	LINE_GRAY
#undef	LINE_GRAY
#endif
#ifdef	LINE_RGB
#undef	LINE_RGB
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#include "stl.h"
#include "line.h"
#include "xyz.h"
#include "cell.h"
#include "sif.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

static int	**face,ID,xL,yT;

static void	Fill(x,y)
int		x,y;
{
	face[yT-y][xL+x]=ID;
}

static Cell	**cell;

static void	Plot(x,y)
int		x,y;
{
	cell[yT-y][xL+x]=1;
}

#define STR_LEN		256
#define MA(num,var)	malloc((size_t)(num)*sizeof(*(var)))

int	main(argc,argv)
int	argc;
char	**argv;
{
	double	scale,x1,y1,z1,x2,y2,z2,dx,dy,dz,
		lo0,la0,dLo,dLa,loD,laD,loR,laR,h,v;
	int	no,size1,y,cnt,idx,Nx,Ny,x,size=0;
	FILE	*file;
	Facet	*facet,*fTop=NULL,**fPtr=(&fTop);
	Line	*line,*lTop=NULL,**lPtr=(&lTop);
	float	*lx,*ly,*lz;
	char	*name,text[STR_LEN],str[6][STR_LEN],
		*fmt,path[STR_LEN],desc[STR_LEN],
		*fnf="%s : file not found.\n",
		*fre="%s : file read error.\n",
		*cmae="cell memory allocation error.";

	if (argc<3) Error("usage : stl_bev_wf STL_files scale");

	if (argv[argc-1][0]!='-') {
	    if ((scale=atof(argv[argc-1]))<=0.0) Error("bad scale.");
	}
	else
	    if ((size=atoi(argv[argc-1]+1))<=0) Error("bad size.");

	while (--argc>1)
	    if (*(name=(*(++argv)))!='-') {
		if ((file=fopen(name,"rb"))==NULL) {
		    fprintf(stderr,fnf,*argv); exit(1);
		}
		if ((*fPtr=ReadFacet(file))==NULL) {
		    fprintf(stderr,fre,*argv); exit(1);
		}
		fclose(file);

		for (facet=(*fPtr); facet->next!=NULL; facet=facet->next) ;

		fPtr=(&(facet->next));
	    }
	    else {
		if ((file=fopen(name+1,"r"))==NULL) {
		    fprintf(stderr,fnf,name); exit(1);
		}
		if ((*lPtr=ReadLine(file))==NULL) {
		    fprintf(stderr,fre,name); exit(1);
		}
		fclose(file);

		for (line=(*lPtr); line->next!=NULL; line=line->next) ;

		 lPtr=(&(line->next));
	    }

	if (fTop==NULL) Error("no facet");

	x1=x2=fTop->x0;
	y1=y2=fTop->y0;
	z1=z2=fTop->z0;
	for (facet=fTop; facet!=NULL; facet=facet->next) {
	    if (facet->x0<x1) x1=facet->x0; if (facet->x0>x2) x2=facet->x0;
	    if (facet->y0<y1) y1=facet->y0; if (facet->y0>y2) y2=facet->y0;
	    if (facet->z0<z1) z1=facet->z0; if (facet->z0>z2) z2=facet->z0;
	    if (facet->x1<x1) x1=facet->x1; if (facet->x1>x2) x2=facet->x1;
	    if (facet->y1<y1) y1=facet->y1; if (facet->y1>y2) y2=facet->y1;
	    if (facet->z1<z1) z1=facet->z1; if (facet->z1>z2) z2=facet->z1;
	    if (facet->x2<x1) x1=facet->x2; if (facet->x2>x2) x2=facet->x2;
	    if (facet->y2<y1) y1=facet->y2; if (facet->y2>y2) y2=facet->y2;
	    if (facet->z2<z1) z1=facet->z2; if (facet->z2>z2) z2=facet->z2;
	}
	for (line=lTop; line!=NULL; line=line->next)
	    for (lx=line->x,
		 ly=line->y,
		 lz=line->z, no=line->No; no>0; no--, lx++, ly++, lz++) {
		if (*lx<x1) x1=(*lx); if (*lx>x2) x2=(*lx);
		if (*ly<y1) y1=(*ly); if (*ly>y2) y2=(*ly);
		if (*lz<z1) z1=(*lz); if (*lz>z2) z2=(*lz);
	    }

	dx=x2-x1;
	dy=y2-y1;
	dz=z2-z1;

	if (size)
	    scale=(double)size/sqrt(dx*dx+dy*dy+dz*dz);
	else
	    size=(int)ceil(scale*sqrt(dx*dx+dy*dy+dz*dz));

	size1=size-1;

	if ((face=(int  **)MA(size,face))==NULL ||
	    (cell=(Cell **)MA(size,cell))==NULL) Error(cmae);

	for (y=0; y<size; y++)
	    if ((face[y]=(int  *)MA(size,*face))==NULL ||
		(cell[y]=(Cell *)MA(size,*cell))==NULL) Error(cmae);

	while (gets(text)!=NULL)
{
	switch(sscanf(text,"%s %s %s %s %s %s",
			   str[0],str[1],str[2],str[3],str[4],str[5])) {
	    case 3:
		lo0=atof(str[0]); dLo=0.0;
		la0=atof(str[1]); dLa=0.0; cnt=1;
		fmt=str[2]; break;
	    case 6:
		lo0=atof(str[0]); dLo=atof(str[2]);
		la0=atof(str[1]); dLa=atof(str[3]); cnt=atoi(str[4]);
		fmt=str[5]; break;
	    default:
		continue;
	}
	for (idx=0; idx<cnt; idx++) {
	    for (y=0; y<size; y++)
	    for (x=0; x<size; x++) {
		face[y][x]=INT_MAX; cell[y][x]=0;
	    }
	    loD=lo0+dLo*(double)idx; loR=loD*M_PI/180.0;
	    laD=la0+dLa*(double)idx; laR=laD*M_PI/180.0;

	    XYZWidth(loR,laR,x1,y1,z1,x2,y2,z2,&h,&v);
	    Nx=(int)ceil(scale*h); xL=(size-Nx)/2;
	    Ny=(int)ceil(scale*v); yT=(size+Ny)/2-1;

	    if (!XYZInit(Nx,Ny,loR,laR,x1,y1,z1,x2,y2,z2))
		Error("rendering memory allocation error.");

	    for (ID=INT_MIN, facet=fTop; facet!=NULL; facet=facet->next, ID++)
		if (XYZSide(loR,laR,
			    facet->x0,facet->y0,facet->z0,
			    facet->x1,facet->y1,facet->z1,
			    facet->x2,facet->y2,facet->z2)>=0.0)
		    XYZFill(facet->x0,facet->y0,facet->z0,
			    facet->x1,facet->y1,facet->z1,
			    facet->x2,facet->y2,facet->z2,Fill);

	    for (y=0; y<size; y++)
	    for (x=0; x<size; x++)
		if ((x>0 && face[y][x]<face[y][x-1]) ||
		    (y>0 && face[y][x]<face[y-1][x]) ||
		    (x<size1 && face[y][x]<face[y][x+1]) ||
		    (y<size1 && face[y][x]<face[y+1][x])) cell[y][x]=1;

	    for (line=lTop; line!=NULL; line=line->next) {
		lx=line->x; ly=line->y; lz=line->z;
		if (line->No==1)
		    XYZPlot((double)*lx,(double)*ly,(double)*lz,Plot);
		else
		    for (no=line->No; no>1; no--, lx++, ly++, lz++)
			XYZDraw((double)lx[0],(double)ly[0],(double)lz[0],
				(double)lx[1],(double)ly[1],(double)lz[1],Plot);
	    }
#ifndef	NF
	    XYZDraw(x1,y1,z1,x2,y1,z1,Plot); XYZDraw(x2,y1,z1,x2,y2,z1,Plot);
	    XYZDraw(x2,y2,z1,x1,y2,z1,Plot); XYZDraw(x1,y2,z1,x1,y1,z1,Plot);
	    XYZDraw(x1,y1,z2,x2,y1,z2,Plot); XYZDraw(x2,y1,z2,x2,y2,z2,Plot);
	    XYZDraw(x2,y2,z2,x1,y2,z2,Plot); XYZDraw(x1,y2,z2,x1,y1,z2,Plot);
	    XYZDraw(x1,y1,z1,x1,y1,z2,Plot); XYZDraw(x2,y1,z1,x2,y1,z2,Plot);
	    XYZDraw(x2,y2,z1,x2,y2,z2,Plot); XYZDraw(x1,y2,z1,x1,y2,z2,Plot);
#endif
	    XYZTerm();

	    sprintf(path,fmt,idx);
	    sprintf(desc,"%f\t%f\n",loD,laD);
	    StoreImageFile(path,size,size,1,cell,desc);
	}
}
	return 0;
}
