
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "stl.h"

#ifndef	SEGMENT
#define SEGMENT		180
#endif

#ifndef	ITERATION
#define ITERATION	1024
#endif

static double	M0,xx=0.0,yy=0.0,zz=0.0,xy=0.0,yz=0.0,zx=0.0;

#define VAL	vec[0]
#define LON	vec[1]
#define LAT	vec[2]

#define LON1	(-M_PI)
#define LON2	M_PI
#define LAT1	0.0
#define LAT2	(M_PI/2.0)

static void	Evaluate(vec)
double		*vec;
{
	double	co,so,co2,so2,cso,ca,sa,ca2,sa2,csa;

	if (LON>LON1 && LON<=LON2 && LAT>=LAT1 && LAT<=LAT2) {
	    co=cos(LON); so=sin(LON); co2=co*co; so2=so*so; cso=co*so; 
	    ca=cos(LAT); sa=sin(LAT); ca2=ca*ca; sa2=sa*sa; csa=ca*sa;
#ifdef	OBLATE
	    VAL=((xx*co2+yy*so2+2.0*xy*cso)*ca2+2.0*(zx*co+yz*so)*csa+zz*sa2);
#else
	    VAL=M0-
		((xx*co2+yy*so2+2.0*xy*cso)*ca2+2.0*(zx*co+yz*so)*csa+zz*sa2);
#endif
	}
	else
	    VAL=M0;
}

static double	can0[3],can1[3],can2[3],*can[3]={can0,can1,can2};

#define STEP	(M_PI/(double)SEGMENT)

#define DEGREE	(180.0/M_PI)

int	main(argc,argv)
int	argc;
char	**argv;
{
	FILE	*file;
	STL	stl;
	double	x,y,z,v,val,lon,lat,vec[3],
		co,so,co2,so2,cso,ca,sa,ca2,sa2,csa,
		C,H2,V2,HV,HmV,rot2,dHV,HV0,b,a,B,A,rot,V=0.0;
	float	f[9];

	if (argc!=2 && argc!=5) {
	    fputs("usage : stl_of STL_file {x0 y0 z0}\n",stderr); exit(1);
	}
	if (argv[1][0]=='-')
	    file=stdin;
	else
	    if ((file=fopen(argv[1],"rb"))==NULL) {
		fprintf(stderr,"%s : STL file not found.\n",argv[1]); exit(1);
	    }

	if (!SetUpSTL(&stl,file)) {
	    fprintf(stderr,"%s : STL header read error.\n",argv[1]); exit(1);
	}

	if (argc==2)
	    x=y=z=0.0;
	else {
	    x=atof(argv[2]); y=atof(argv[3]); z=atof(argv[4]);
	}

	while (ScanSTL(&stl,f+0,f+1,f+2,f+3,f+4,f+5,f+6,f+7,f+8)>0) {

#define x0	(double)f[0]
#define y0	(double)f[1]
#define z0	(double)f[2]
#define x1	(double)f[3]
#define y1	(double)f[4]
#define z1	(double)f[5]
#define x2	(double)f[6]
#define y2	(double)f[7]
#define z2	(double)f[8]

	    V+=(v=x0*y1*z2+y0*z1*x2+z0*x1*y2-z0*y1*x2-y0*x1*z2-x0*z1*y2);

	    if (argc==2) {
		x+=(v*(x0+x1+x2));
		y+=(v*(y0+y1+y2));
		z+=(v*(z0+z1+z2));
	    }
	    xx+=(v*(x0*x0+x1*x1+x2*x2+x0*x1+x1*x2+x2*x0));
	    yy+=(v*(y0*y0+y1*y1+y2*y2+y0*y1+y1*y2+y2*y0));
	    zz+=(v*(z0*z0+z1*z1+z2*z2+z0*z1+z1*z2+z2*z0));
	    xy+=(v*(x0*y0+x1*y1+x2*y2+
		   (x0*y1+x1*y2+x2*y0+x0*y2+x1*y0+x2*y1)/2.0));
	    yz+=(v*(y0*z0+y1*z1+y2*z2+
		   (y0*z1+y1*z2+y2*z0+y0*z2+y1*z0+y2*z1)/2.0));
	    zx+=(v*(z0*x0+z1*x1+z2*x2+
		   (z0*x1+z1*x2+z2*x0+z0*x2+z1*x0+z2*x1)/2.0));

	}
	if (V<=0.0) {
	    fputs("non-positive total volume.\n",stderr); exit(1);
	}
	V/=6.0;

	if (argc==2) {
	    x/=(24.0*V);
	    y/=(24.0*V);
	    z/=(24.0*V);
	}
	xx=xx/60.0-x*x*V;
	yy=yy/60.0-y*y*V;
	zz=zz/60.0-z*z*V;
	xy=xy/60.0-x*y*V;
	yz=yz/60.0-y*z*V;
	zx=zx/60.0-z*x*V;

	val=M0=xx+yy+zz; lon=LON2; lat=LAT2;
	for (LON=LON2; LON> LON1; LON-=STEP)
	for (LAT=LAT2; LAT>=LAT1; LAT-=STEP) {
	    Evaluate(vec);
	    if (VAL<val) {
		val=VAL; lon=LON; lat=LAT;
	    }
	}
	can0[1]=can2[1]=lon; can1[1]=lon+STEP;
	can0[2]=can1[2]=lat; can2[2]=lat-STEP;

	Simplex(2,can,Evaluate,ITERATION,vec);

	lon=LON;
	lat=LAT;
#ifdef	OBLATE
	C=sqrt(5.0*VAL/V);
#else
	C=sqrt(5.0*(M0-VAL)/V);
#endif
	co=cos(lon); so=sin(lon); co2=co*co; so2=so*so; cso=co*so;
	ca=cos(lat); sa=sin(lat); ca2=ca*ca; sa2=sa*sa; csa=ca*sa;
	H2=xx*so2+yy*co2-2.0*xy*cso;
	V2=(xx*co2+yy*so2+2.0*xy*cso)*sa2-2.0*(zx*co+yz*so)*csa+zz*ca2;
	HV=((xx-yy)*cso-xy*(co2-so2))*sa-(zx*so-yz*co)*ca;
	HmV=(H2-V2)/2.0;
	rot2=atan2(HV,HmV);
	dHV=HmV*cos(rot2)+HV*sin(rot2);
	HV0=(H2+V2)/2.0;

#ifdef	OBLATE
	if (dHV>=0.0) {
#else
	if (dHV<=0.0) {
#endif
	    b=HV0-dHV; a=HV0+dHV; if (rot2<0.0) rot2+=(M_PI*2.0);
	}
	else {
	    b=HV0+dHV; a=HV0-dHV; rot2+=M_PI;
	}
	B=sqrt(5.0*b/V); A=sqrt(5.0*a/V); rot=rot2/2.0;

	printf("%lf\n%lg\t%lg\t%lg\n%lg\t%lg\t%lg\n%lg\t%lg\t%lg\n",
	       V,x,y,z,lon*DEGREE,lat*DEGREE,rot*DEGREE,A,B,C);

	return 0;
}
