
#ifndef	STL_COLOR
#define STL_COLOR
#endif

#include <stdio.h>
#include <stdlib.h>
#include "stl.h"

#ifndef	HASH_BITS
#define HASH_BITS	4
#endif

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

#define HASH_SIZE	(1<<HASH_BITS)

static int	Hash(xyz)
unsigned char	*xyz;
{
	return ((int)xyz[0]+(int)xyz[1]+(int)xyz[2]+(int)xyz[3])&(HASH_SIZE-1);
}

#define HASH(x,y,z)	x*x+y*y+z*z

typedef	struct Vertex {
		float		x,y,z;
		union {
			float	hash;
			int	name;
		} u;
		struct Vertex	*next;
	} Vertex;

static Vertex	*table[HASH_SIZE][HASH_SIZE][HASH_SIZE];

static Vertex	*Register(x,y,z)
float		x,y,z;
{
	Vertex	*vertex,**top=(&(table[Hash(&x)][Hash(&y)][Hash(&z)]));
	float	hash=HASH(x,y,z);

	for (vertex=(*top); vertex!=NULL; vertex=vertex->next)
	    if (hash==vertex->u.hash && x==vertex->x &&
					y==vertex->y &&
					z==vertex->z) return vertex;

	if ((vertex=(Vertex *)malloc(sizeof(*vertex)))==NULL)
	    Error("vertex memory allocation error.");

	vertex->x=x;
	vertex->y=y;
	vertex->z=z;
	vertex->u.hash=hash;
	vertex->next=(*top); return *top=vertex;
}

static void	PutNormal(file,ptr)
FILE		*file;
char		*ptr;
{
	fwrite(ptr,1,4,file);
}

static void	PutReverse(file,ptr)
FILE		*file;
unsigned char	*ptr;
{
fputc(ptr[3],file); fputc(ptr[2],file); fputc(ptr[1],file); fputc(ptr[0],file);
}

typedef struct Face {
		Vertex		*vertex[3];
		short		flag;
		struct Face	*next;
	} Face;

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	x,y,z,faces,vertices,R=255,G=255,B=255;
	FILE	*file;
	STL	stl;
	float	x0,y0,z0,x1,y1,z1,x2,y2,z2;
	short	flag;
	Face	*face,*top=NULL,**ptr=(&top);
	Vertex	*vertex;

	short	s=0x1234;
        char	*c=(char *)&s;
	void	(*Put)()=(*c==(char)(s&0xff))?PutNormal:PutReverse;

	if (argc!=3 && argc!=6) Error("usage : stl_zcp STL {R G B} ZCP");

	if (argc==6)
	    if ((R=atoi(argv[2]))<0 || R>255 ||
		(G=atoi(argv[3]))<0 || G>255 ||
		(B=atoi(argv[4]))<0 || B>255) Error("bad surface color.");

	if (argv[1][0]=='-')
	    file=stdin;
	else
	    if ((file=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if (!SetUpSTL(&stl,file)) Error("header read error.");

	for (x=0; x<HASH_SIZE; x++)
	for (y=0; y<HASH_SIZE; y++)
	for (z=0; z<HASH_SIZE; z++) table[x][y][z]=NULL;

	for (faces=0; ScanSTL(&stl,&x0,&y0,&z0,
				   &x1,&y1,&z1,
				   &x2,&y2,&z2,&flag)>0; faces++) {
	    if ((face=(Face *)malloc(sizeof(*face)))==NULL)
		Error("face memory allocation error.");

	    face->vertex[0]=Register(x0,y0,z0);
	    face->vertex[1]=Register(x1,y1,z1);
	    face->vertex[2]=Register(x2,y2,z2);
	    face->flag=flag;
	    face->next=NULL; *ptr=face; ptr=(&(face->next));
	}
	if (file!=stdin) fclose(file);

	vertices=0;
	for (x=0; x<HASH_SIZE; x++)
	for (y=0; y<HASH_SIZE; y++)
	for (z=0; z<HASH_SIZE; z++)
	    for (vertex=table[x][y][z]; vertex!=NULL; vertex=vertex->next)
		vertex->u.name=vertices++;

	if (argv[argc-1][0]=='-')
	    file=stdout;
	else
	    if ((file=fopen(argv[argc-1],"wb"))==NULL)
		Error("file not stored.");

	fprintf(file,"ply\n"
		     "format binary_little_endian 1.0\n"
		     "element vertex %d\n"
		     "property float x\n"
		     "property float y\n"
		     "property float z\n"
		     "element face %d\n"
		     "property uchar red\n"
		     "property uchar green\n"
		     "property uchar blue\n"
		     "property list uchar int vertex_indices\n"
		     "end_header\n",
                     vertices,faces);

	for (x=0; x<HASH_SIZE; x++)
	for (y=0; y<HASH_SIZE; y++)
	for (z=0; z<HASH_SIZE; z++)
	    for (vertex=table[x][y][z]; vertex!=NULL; vertex=vertex->next) {
		(*Put)(file,&(vertex->x));
		(*Put)(file,&(vertex->y));
		(*Put)(file,&(vertex->z));
	    }

	for (face=top; face!=NULL; face=face->next) {
	    if (STL_COLOR_FLAG(face->flag)) {
		fputc((int)(STL_COLOR_R(face->flag)*255.0+0.5),file);
		fputc((int)(STL_COLOR_G(face->flag)*255.0+0.5),file);
		fputc((int)(STL_COLOR_B(face->flag)*255.0+0.5),file);
	    }
	    else {
		fputc(R,file);
		fputc(G,file);
		fputc(B,file);
	    }
	    fputc(3,file); (*Put)(file,&(face->vertex[0]->u.name));
			   (*Put)(file,&(face->vertex[1]->u.name));
			   (*Put)(file,&(face->vertex[2]->u.name));
	}
	if (file!=stdout) fclose(file);

	return 0;
}
