
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "csi.h"
#include "sgif.h"

#if	defined(CSI_RGB)
#include "rif_rgb.h"
#elif	defined(T2G_CM)
#include "rif_cm.h"
#else
#include "rif.h"
#endif

#define BYTE	unsigned char

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	**path;
	int	Nx,Ny,Nz,BPS,z,nx,ny,y,x;
	FILE	*file;

	if (argc!=4) {
	    fputs("usage : t2g_movie directory nameFile GIF\n",stderr); exit(1);
	}

#if	defined(CSI_RGB)
	path=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,NULL);
#else
	path=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,&BPS);

#if	defined(T2G_CM)
	if (BPS>8) {
	    fputs("too many bits per sample.\n",stderr); exit(1);
	}
#endif
#endif
	if (argv[3][0]=='-' && argv[3][1]=='\0')
	    file=stdout;
	else
	    if ((file=fopen(argv[3],"wb"))==NULL) {
		fprintf(stderr,"%s : file not stored.\n",argv[3]); exit(1);
	    }

#if	defined(CSI_RGB) || defined(T2G_CM)
	InitStoreGIF(file,Nx,Ny,NULL,NULL,NULL);

#if	defined(CSI_RGB)
{
	Cell	**R,**G,**B;
	BYTE	**r,**g,**b;
	double	ratio;

	for (z=0; z<Nz; z++) {
	    ReadImageFile_RGB(path[z],&nx,&ny,&BPS,&R,&G,&B,NULL);
	    if (nx>Nx || ny>Ny) {
		fprintf(stderr,"%s : bad image size.\n",path[z]); exit(1);
	    }
	    r=(BYTE **)R;
	    g=(BYTE **)G;
	    b=(BYTE **)B; ratio=255.0/(double)((1<<BPS)-1);
	    for (y=0; y<ny; y++)
	    for (x=0; x<nx; x++) {
		r[y][x]=(BYTE)((double)R[y][x]*ratio+0.5);
		b[y][x]=(BYTE)((double)B[y][x]*ratio+0.5);
		g[y][x]=(BYTE)((double)G[y][x]*ratio+0.5);
	    }
	    StoreGIF_RGB(file,(Nx-nx)/2,(Ny-ny)/2,nx,ny,r,g,b);

	    for (y=ny-1; y>=0; y--) {
		free(B[y]); free(G[y]); free(R[y]);
	    }
	    free(B); free(G); free(R);
	}
}
#else
{
	unsigned short	*R,*G,*B;
	Cell		**PV;
	int		index;
	BYTE		r[256],g[256],b[256],**pv;

	for (z=0; z<Nz; z++) {
	    ReadImageFile_CM(path[z],&nx,&ny,&BPS,&R,&G,&B,&PV,NULL);
	    if (nx>Nx || ny>Ny) {
		fprintf(stderr,"%s : bad image size.\n",path[z]); exit(1);
	    }
	    if (BPS>8) {
		fprintf(stderr,"%s : bad bits per sample.\n",path[z]); exit(1);
	    }
	    for (index=(1<<BPS)-1; index>=0; index--) {
		r[index]=(BYTE)(R[index]>>8);
		g[index]=(BYTE)(G[index]>>8);
		b[index]=(BYTE)(B[index]>>8);
	    }
	    free(R); free(G); free(B);

	    pv=(BYTE **)PV;
	    for (y=0; y<ny; y++)
	    for (x=0; x<nx; x++) pv[y][x]=(BYTE)PV[y][x];

	    StoreGIF(file,(Nx-nx)/2,(Ny-ny)/2,nx,ny,r,g,b,pv);

	    for (y=ny-1; y>=0; y--) free(PV[y]); free(PV);
	}
}
#endif
#else
	InitStoreGIF(file,Nx,Ny,StoreGIF_Gray,StoreGIF_Gray,StoreGIF_Gray);
{
	Cell	**PV;
	BYTE	**pv;
	double	ratio=255.0/(double)((1<<BPS)-1);

	for (z=0; z<Nz; z++) {
	    ReadImageFile(path[z],&nx,&ny,NULL,&PV,NULL);
	    if (nx>Nx || ny>Ny) {
		fprintf(stderr,"%s : bad image size.\n",path[z]); exit(1);
	    }
	    pv=(BYTE **)PV;
	    for (y=0; y<ny; y++)
	    for (x=0; x<nx; x++) pv[y][x]=(BYTE)((double)PV[y][x]*ratio+0.5);

	    StoreGIF(file,(Nx-nx)/2,(Ny-ny)/2,nx,ny,NULL,NULL,NULL,pv);

	    for (y=ny-1; y>=0; y--) free(PV[y]); free(PV);
	}
}
#endif
	TermStoreGIF(file); if (file!=stdout) fclose(file); return 0;
}
