
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

typedef unsigned short	Depth;

static Depth	**depth;
static int	Ni,Nj,*ijMin,*ijMax;
static double	hx,hy,vx,vy,vz,dx,dy,dz,h0,v0,d0,Sh,Sv,Sd;

#define H(x,y)		(hx*(x)+hy*(y))
#define V(x,y,z)	(vx*(x)+vy*(y)+vz*(z))
#define D(x,y,z)	(dx*(x)+dy*(y)+dz*(z))

#define ALLOC(type,noe)	(type *)malloc(sizeof(type)*(size_t)(noe))

int	XYZInit(iSize,jSize,lo,la,xMin,yMin,zMin,xMax,yMax,zMax)
int	iSize,jSize;
double	lo,la,xMin,yMin,zMin,xMax,yMax,zMax;
{
	int	Nij,i,j;
	double	sLo,cLo,sLa,cLa,h,v,d;

	Ni=iSize;
	Nj=jSize; Nij=(iSize>jSize)?iSize:jSize;
	if ((ijMin=ALLOC(int,Nij))==NULL ||
	    (ijMax=ALLOC(int,Nij))==NULL ||
	    (depth=ALLOC(Depth *,Nj))==NULL) return 0;

	for (j=0; j<Nj; j++) {
	    if ((depth[j]=ALLOC(Depth,Ni))==NULL) {
		while (--j>=0) free(depth[j]); free(depth);
		free(ijMax); free(ijMin);
		return 0;
	    }
	    for (i=0; i<Ni; i++) depth[j][i]=0;
	}
	sLo=sin(lo); cLo=cos(lo); hx=(-sLo);     hy=  cLo;
	sLa=sin(la); cLa=cos(la); vx=(-cLo*sLa); vy=(-sLo*sLa); vz=cLa;
				  dx=  cLo*cLa;  dy=  sLo*cLa;  dz=sLa;
	h0=Sh=H(xMin,yMin);
	if ((h=H(xMin,yMax))<h0) h0=h; else if (h>Sh) Sh=h;
	if ((h=H(xMax,yMin))<h0) h0=h; else if (h>Sh) Sh=h;
	if ((h=H(xMax,yMax))<h0) h0=h; else if (h>Sh) Sh=h;
	Sh=(double)(Ni-1)/(Sh-h0);
	v0=Sv=V(xMin,yMin,zMin);
	if ((v=V(xMin,yMin,zMax))<v0) v0=v; else if (v>Sv) Sv=v;
	if ((v=V(xMin,yMax,zMin))<v0) v0=v; else if (v>Sv) Sv=v;
	if ((v=V(xMin,yMax,zMax))<v0) v0=v; else if (v>Sv) Sv=v;
	if ((v=V(xMax,yMin,zMin))<v0) v0=v; else if (v>Sv) Sv=v;
	if ((v=V(xMax,yMin,zMax))<v0) v0=v; else if (v>Sv) Sv=v;
	if ((v=V(xMax,yMax,zMin))<v0) v0=v; else if (v>Sv) Sv=v;
	if ((v=V(xMax,yMax,zMax))<v0) v0=v; else if (v>Sv) Sv=v;
	Sv=(double)(Nj-1)/(Sv-v0);
	d0=Sd=D(xMin,yMin,zMin);
	if ((d=D(xMin,yMin,zMax))<d0) d0=d; else if (d>Sd) Sd=d;
	if ((d=D(xMin,yMax,zMin))<d0) d0=d; else if (d>Sd) Sd=d;
	if ((d=D(xMin,yMax,zMax))<d0) d0=d; else if (d>Sd) Sd=d;
	if ((d=D(xMax,yMin,zMin))<d0) d0=d; else if (d>Sd) Sd=d;
	if ((d=D(xMax,yMin,zMax))<d0) d0=d; else if (d>Sd) Sd=d;
	if ((d=D(xMax,yMax,zMin))<d0) d0=d; else if (d>Sd) Sd=d;
	if ((d=D(xMax,yMax,zMax))<d0) d0=d; else if (d>Sd) Sd=d;
	Sd=(pow(2.0,(double)(8*sizeof(Depth)))-1.0)/(Sd-d0);
	return 1;
}

void	XYZTerm()
{
	int	j;

	for (j=Nj-1; j>=0; j--) free(depth[j]); free(depth);
	free(ijMax); free(ijMin);
}

static void	(*Store)();

static void	Test(i,j,k)
int		i,j,k;
{
	double	h,v,d;

	if (k>(int)depth[j][i]) {
	    depth[j][i]=(Depth)k;
	    if (Store!=NULL) {
		h=h0+(double)i/Sh; v=v0+(double)j/Sv; d=d0+(double)k/Sd;
		(*Store)(i,j,h*hx+v*vx+d*dx,h*hy+v*vy+d*dy,v*vz+d*dz);
	    }
	}
}

#define I(x,y)		((int)(Sh*(H(x,y  )-h0)+0.5))
#define J(x,y,z)	((int)(Sv*(V(x,y,z)-v0)+0.5))
#define K(x,y,z)	((int)(Sd*(D(x,y,z)-d0)+0.5))

void	XYZPlot(x,y,z,store)
double	x,y,z;
void	(*store)();
{
	Store=store; Test(I(x,y),J(x,y,z),K(x,y,z));
}

void	XYZDraw(x1,y1,z1,x2,y2,z2,store)
double	x1,y1,z1,x2,y2,z2;
void	(*store)();
{
	int	di,dj,si,sj,m,e,p,i1=I(x1,y1),j1=J(x1,y1,z1),k1=K(x1,y1,z1),
				  i2=I(x2,y2),j2=J(x2,y2,z2),k2=K(x2,y2,z2);
	double	dk,k0;

	Store=store;
	if (i1==i2 && j1==j2)
	    if (k1>k2) Test(i1,j1,k1); else Test(i2,j2,k2);
	else {
	    dk=(double)(k2-k1); k0=(double)k1+0.5;
	    if (i2>i1) { di=i2-i1; si=1; } else { di=i1-i2; si=(-1); }
	    if (j2>j1) { dj=j2-j1; sj=1; } else { dj=j1-j2; sj=(-1); }
	    if (di>dj)
		for (m=2*dj, e=m-di, p=m-2*di,
		     dk/=(double)di, i2+=si; i1!=i2; i1+=si, k0+=dk) {
		    Test(i1,j1,(int)k0); if (e<0) e+=m; else { e+=p; j1+=sj; }
		}
	    else
		for (m=2*di, e=m-dj, p=m-2*dj,
		     dk/=(double)dj, j2+=sj; j1!=j2; j1+=sj, k0+=dk) {
		    Test(i1,j1,(int)k0); if (e<0) e+=m; else { e+=p; i1+=si; }
		}
	}
}

static void	Scan(i1,j1,i2,j2)
int		i1,j1,i2,j2;
{
	int	di,dj,si,sj,m,e,p,*min=ijMin+i1,*max=ijMax+i1;

	if (i2>i1) { di=i2-i1; si=1; } else { di=i1-i2; si=(-1); }
	if (j2>j1) { dj=j2-j1; sj=1; } else { dj=j1-j2; sj=(-1); }
	if (di>dj)
	    for (m=2*dj, e=m-di, p=m-2*di; i1!=i2; i1+=si) {
		if (j1<*min) *min=j1; if (j1>*max) *max=j1;
		if (e<0) e+=m; else { e+=p; j1+=sj; } min+=si; max+=si;
	    }
	else
	    for (m=2*di, e=m-dj, p=m-2*dj; j1!=j2; j1+=sj) {
		if (j1<*min) *min=j1; if (j1>*max) *max=j1;
		if (e<0) e+=m; else { e+=p; i1+=si; min+=si; max+=si; }
	    }
}

void	XYZFill(x1,y1,z1,x2,y2,z2,x3,y3,z3,store)
double	x1,y1,z1,x2,y2,z2,x3,y3,z3;
void	(*store)();
{
	int	dk,j23,j31,j12,i32,i13,i21,iMax,iMin,jMax,jMin,i,j,
		i1=I(x1,y1),j1=J(x1,y1,z1),
		i2=I(x2,y2),j2=J(x2,y2,z2),
		i3=I(x3,y3),j3=J(x3,y3,z3);
	double	k0,k1,k2,k3,ki,kj,k;

	if ((dk=i1*j2+i2*j3+i3*j1-i3*j2-i2*j1-i1*j3)==0) return;

	Store=store;
	k0=((double)(i2*j3-i3*j2)*(k1=(double)K(x1,y1,z1))+
	    (double)(i3*j1-i1*j3)*(k2=(double)K(x2,y2,z2))+
	    (double)(i1*j2-i2*j1)*(k3=(double)K(x3,y3,z3)))/(double)dk+0.5;
	ki=((double)(j23=j2-j3)*k1+
	    (double)(j31=j3-j1)*k2+
	    (double)(j12=j1-j2)*k3)/(double)dk;
	kj=((double)(i32=i3-i2)*k1+
	    (double)(i13=i1-i3)*k2+
	    (double)(i21=i2-i1)*k3)/(double)dk;
	if (dk<0) {
	    j23=(-j23); j31=(-j31); j12=(-j12);
	    i32=(-i32); i13=(-i13); i21=(-i21);
	}
	if ((iMax=(i1>i2)?(i1>i3)?i1:i3:(i2>i3)?i2:i3)-
	    (iMin=(i1<i2)?(i1<i3)?i1:i3:(i2<i3)?i2:i3)>
	    (jMax=(j1>j2)?(j1>j3)?j1:j3:(j2>j3)?j2:j3)-
	    (jMin=(j1<j2)?(j1<j3)?j1:j3:(j2<j3)?j2:j3)) {
	    for (i=iMin; i<=iMax; i++) { ijMin[i]=jMax; ijMax[i]=jMin; }
	    Scan(i1,j1,i2,j2); Scan(i2,j2,i3,j3); Scan(i3,j3,i1,j1);
	    for (k0+=(ki*(double)(i=iMin)),
		 i1=j12*(i-i1)-i21*j1,
		 i2=j23*(i-i2)-i32*j2,
		 i3=j31*(i-i3)-i13*j3; i<=iMax; i1+=j12,
					        i2+=j23,
						i3+=j31, k0+=ki, i++)
		for (k=k0+kj*(double)(j=ijMin[i]),
		     j1=i1+i21*j,
		     j2=i2+i32*j,
		     j3=i3+i13*j; j<=ijMax[i]; j1+=i21,
					       j2+=i32,
					       j3+=i13, k+=kj, j++)
		    if (j1>=0 && j2>=0 && j3>=0) Test(i,j,(int)k);
	}
	else {
	    for (j=jMin; j<=jMax; j++) { ijMin[j]=iMax; ijMax[j]=iMin; }
	    Scan(j3,i3,j2,i2); Scan(j2,i2,j1,i1); Scan(j1,i1,j3,i3);
	    for (k0+=(kj*(double)(j=jMin)),
		 j1=i21*(j-j1)-j12*i1,
		 j2=i32*(j-j2)-j23*i2,
		 j3=i13*(j-j3)-j31*i3; j<=jMax; j1+=i21,
					        j2+=i32,
						j3+=i13, k0+=kj, j++)
		for (k=k0+ki*(double)(i=ijMin[j]),
		     i1=j1+j12*i,
		     i2=j2+j23*i,
		     i3=j3+j31*i; i<=ijMax[j]; i1+=j12,
					       i2+=j23,
					       i3+=j31, k+=ki, i++)
		    if (i1>=0 && i2>=0 && i3>=0) Test(i,j,(int)k);
	}
}

double	XYZSide(lo,la,x1,y1,z1,x2,y2,z2,x3,y3,z3)
double	lo,la,x1,y1,z1,x2,y2,z2,x3,y3,z3;
{
	double	sLo=sin(lo),cLo=cos(lo),
		sLa=sin(la),cLa=cos(la),
		dx12=x2-x1,dy12=y2-y1,dz12=z2-z1,
		dx13=x3-x1,dy13=y3-y1,dz13=z3-z1,
		nx=dy12*dz13-dz12*dy13,
		ny=dz12*dx13-dx12*dz13,
		nz=dx12*dy13-dy12*dx13,n=nx*nx+ny*ny+nz*nz;

	return (n==0.0)?0.0:((nx*cLo+ny*sLo)*cLa+nz*sLa)/sqrt(n);
}

#define A(x,y)		(-x*sLo+y*cLo)
#define B(x,y,z)	(-x*cLo-y*sLo)*sLa+z*cLa

void	XYZWidth(lo,la,xMin,yMin,zMin,xMax,yMax,zMax,hWidth,vWidth)
double	lo,la,xMin,yMin,zMin,xMax,yMax,zMax,*hWidth,*vWidth;
{
	double	sLo=sin(lo),cLo=cos(lo),
		sLa=sin(la),cLa=cos(la),h1,h2,v1,v2,h,v;

	h1=h2=A(xMin,yMin);
	if ((h=A(xMin,yMax))<h1) h1=h; else if (h>h2) h2=h;
	if ((h=A(xMax,yMin))<h1) h1=h; else if (h>h2) h2=h;
	if ((h=A(xMax,yMax))<h1) h1=h; else if (h>h2) h2=h;
	*hWidth=h2-h1;
	v1=v2=B(xMin,yMin,zMin);
	if ((v=B(xMin,yMin,zMax))<v1) v1=v; else if (v>v2) v2=v;
	if ((v=B(xMin,yMax,zMin))<v1) v1=v; else if (v>v2) v2=v;
	if ((v=B(xMin,yMax,zMax))<v1) v1=v; else if (v>v2) v2=v;
	if ((v=B(xMax,yMin,zMin))<v1) v1=v; else if (v>v2) v2=v;
	if ((v=B(xMax,yMin,zMax))<v1) v1=v; else if (v>v2) v2=v;
	if ((v=B(xMax,yMax,zMin))<v1) v1=v; else if (v>v2) v2=v;
	if ((v=B(xMax,yMax,zMax))<v1) v1=v; else if (v>v2) v2=v;
	*vWidth=v2-v1;
}

double	XYZRatio(lo,la,xMin,yMin,zMin,xMax,yMax,zMax)
double	lo,la,xMin,yMin,zMin,xMax,yMax,zMax;
{
	double	h,v;

	XYZWidth(lo,la,xMin,yMin,zMin,xMax,yMax,zMax,&h,&v); return v/h;
}

void	XYZShade(lo,la,store)
double	lo,la;
void	(*store)();
{
	double	lx,ly,lz,dh,dv,nx,ny,nz,ip;
	int	i,j,Ni1=Ni-1,Nj1=Nj-1;

	if (Ni1<=0 || Nj1<=0) return;

	lx=cos(lo)*cos(la); ly=sin(lo)*cos(la); lz=sin(la);
	for (j=0; j<=Nj1; j++)
	for (i=0; i<=Ni1; i++)
	    if (depth[j][i]>0) {
		dh=((i==0)?(double)depth[j][i+1]:
		    (i==Ni1)?-(double)depth[j][i-1]:
		    ((double)depth[j][i+1]-(double)depth[j][i-1])/2.0)*Sh/Sd;
		dv=((j==0)?(double)depth[j+1][i]:
		    (j==Nj1)?-(double)depth[j-1][i]:
		    ((double)depth[j+1][i]-(double)depth[j-1][i])/2.0)*Sv/Sd;
		(*store)(i,j,((ip=(nx=dx-dh*hx-dv*vx)*lx+
				  (ny=dy-dh*hy-dv*vy)*ly+(nz=dz-dv*vz)*lz)>0.0)?
			     ip/sqrt(nx*nx+ny*ny+nz*nz):0.0);
	    }
}
