
#include <stdio.h>
#include <stdlib.h>
#include <string.h>	/* using strcmp() */

static void	Error(char *msg)
{
	(void)fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

static int	Fread4_N(char *ptr,FILE *file)
{
	return fread(ptr,(size_t)4,(size_t)1,file)==1;
}

static int	Fread4_R(char *ptr,FILE *file)
{
	return fread(ptr+3,(size_t)1,(size_t)1,file)==1 &&
	       fread(ptr+2,(size_t)1,(size_t)1,file)==1 &&
	       fread(ptr+1,(size_t)1,(size_t)1,file)==1 &&
	       fread(ptr+0,(size_t)1,(size_t)1,file)==1;
}

static void	Fwrite4_N(char *ptr,FILE *file)
{
	(void)fwrite(ptr,(size_t)4,(size_t)1,file);
}

static void	Fwrite4_R(char *ptr,FILE *file)
{
	(void)fwrite(ptr+3,(size_t)1,(size_t)1,file);
	(void)fwrite(ptr+2,(size_t)1,(size_t)1,file);
	(void)fwrite(ptr+1,(size_t)1,(size_t)1,file);
	(void)fwrite(ptr+0,(size_t)1,(size_t)1,file);
}

int	main(int argc,char **argv)
{
	FILE		*org,*new;
	unsigned	V,F,v,f,v0,v1,v2;
	short		word=127;
	char		e_h[12],rgb3[4],
			*byte=(char *)&word;
	int		w_b=(char)(word&255)==(*byte),
			(*Fread4)()=(w_b)?Fread4_N:Fread4_R;
	void		(*Fwrite4)()=(w_b)?Fwrite4_N:Fwrite4_R;
	double		Sx,Sy,Sz,x1,y1,z1,x2,y2,z2;
	float		x,y,z;

	if (argc!=2 && argc!=4 && argc!=6)
	    Error("usage : zcp_scale orgZCP {scale|(Sx Sy Sz) newZCP}");

	if (argv[1][0]=='-')
	    org=stdin;
	else
	    if ((org=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if (fscanf(org,"ply\n"
		       "format binary_little_endian 1.0\n"
		       "element vertex %u\n"
		       "property float x\n"
		       "property float y\n"
		       "property float z\n"
		       "element face %u\n"
		       "property uchar red\n"
		       "property uchar green\n"
		       "property uchar blue\n"
		       "property list uchar int vertex_indices\n",&V,&F)!=2 ||
	    fgets(e_h,12,org)==NULL || strcmp(e_h,"end_header\n") ||
	    V==0 || F==0) Error("bad header.");

	if (argc!=2) {
	    if (argc==4)
		Sx=Sy=Sz=atof(argv[2]);
	    else {
		Sx=atof(argv[2]);
		Sy=atof(argv[3]);
		Sz=atof(argv[4]);
	    }
	    if (argv[argc-1][0]=='-')
		new=stdout;
	    else
		if ((new=fopen(argv[argc-1],"wb"))==NULL)
		    Error("file not stored.");

	    (void)fprintf(new,"ply\n"
			      "format binary_little_endian 1.0\n"
			      "element vertex %u\n"
			      "property float x\n"
			      "property float y\n"
			      "property float z\n"
			      "element face %u\n"
			      "property uchar red\n"
			      "property uchar green\n"
			      "property uchar blue\n"
			      "property list uchar int vertex_indices\n"
			      "end_header\n",V,F);
	}
	for (v=0; v<V; v++) {
	    if (!(*Fread4)((char *)&x,org) ||
		!(*Fread4)((char *)&y,org) ||
		!(*Fread4)((char *)&z,org)) Error("bad vertex.");

	    if (argc!=2) {
		x*=Sx; (*Fwrite4)((char *)&x,new);
		y*=Sy; (*Fwrite4)((char *)&y,new);
		z*=Sz; (*Fwrite4)((char *)&z,new);
	    }
	    else
		if (v==0) {
		    x1=x2=x;
		    y1=y2=y;
		    z1=z2=z;
		}
		else {
		    if (x<x1) x1=x; else if (x>x2) x2=x;
		    if (y<y1) y1=y; else if (y>y2) y2=y;
		    if (z<z1) z1=z; else if (z>z2) z2=z;
		}
	}
	if (argc==2)
	    (void)printf("%lf\t%lf\t%lf\n"
			 "%lf\t%lf\t%lf\n"
			 "%lf\t%lf\t%lf\n",x1,x2,x2-x1,
					   y1,y2,y2-y1,
					   z1,z2,z2-z1);

	for (f=0; f<F; f++) {
	    if (fread(rgb3,(size_t)1,(size_t)4,org)!=4 ||
		rgb3[3]!=3 ||
		!(*Fread4)((char *)&v0,org) || v0>=V ||
		!(*Fread4)((char *)&v1,org) || v1>=V ||
		!(*Fread4)((char *)&v2,org) || v2>=V) Error("bad face.");

	    if (argc!=2) {
		(void)fwrite(rgb3,(size_t)1,(size_t)4,new);
		(*Fwrite4)((char *)&v0,new);
		(*Fwrite4)((char *)&v1,new);
		(*Fwrite4)((char *)&v2,new);
	    }
	}
	if (argc!=2 && new!=stdout) (void)fclose(new);

	if (org!=stdin) (void)fclose(org); return 0;
}
